#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from xcube.constants import EXTENSION_POINT_DATA_STORES
from xcube.util import extension

from .constants import DATA_STORE_ID


def init_plugin(ext_registry: extension.ExtensionRegistry):
    ext_registry.add_extension(
        loader=extension.import_component("xcube_eopf.store:EOPFZarrDataStore"),
        point=EXTENSION_POINT_DATA_STORES,
        name=DATA_STORE_ID,
        description="EOPF-Zarr DataStore",
    )



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from collections.abc import Container, Iterator
from typing import Any

import pystac_client
import xarray as xr
from xcube.core.mldataset import MultiLevelDataset
from xcube.core.store import (
    DATASET_TYPE,
    DatasetDescriptor,
    DataStore,
    DataStoreError,
    DataTypeLike,
)
from xcube.util.jsonschema import JsonObjectSchema

from .constants import EOPF_ZARR_OPENR_ID, STAC_URL
from .prodhandler import ProductHandler
from .prodhandlers import register_product_handlers
from .utils import (
    bbox_to_geojson,
    filter_items_deprecated,
    filter_items_wrong_footprint,
    reproject_bbox,
)


class EOPFZarrDataStore(DataStore):
    """EOPF-Zarr implementation of the data store."""

    def __init__(self):
        register_product_handlers()

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            description="Describes the parameters of the xcube data store 'eopf-zarr'.",
            properties=dict(),
            required=[],
            additional_properties=False,
        )

    @classmethod
    def get_data_types(cls) -> tuple[str, ...]:
        return (DATASET_TYPE.alias,)

    def get_data_types_for_data(self, data_id: str) -> tuple[str, ...]:
        self._assert_has_data(data_id)
        return (DATASET_TYPE.alias,)

    def get_data_ids(
        self,
        data_type: DataTypeLike = None,
        include_attrs: Container[str] | bool = False,
    ) -> Iterator[str | tuple[str, dict[str, Any]], None]:
        self._assert_valid_data_type(data_type)
        for data_id in ProductHandler.registry.keys():
            if not include_attrs:
                yield data_id
            else:
                yield data_id, dict()

    def has_data(self, data_id: str, data_type: DataTypeLike = None) -> bool:
        self._assert_valid_data_type(data_type)
        if data_id in ProductHandler.registry.keys():
            return True
        return False

    def get_data_opener_ids(
        self, data_id: str = None, data_type: DataTypeLike = None
    ) -> tuple[str, ...]:
        self._assert_valid_data_type(data_type)
        if data_id is not None:
            self._assert_has_data(data_id)
        return (EOPF_ZARR_OPENR_ID,)

    def get_open_data_params_schema(
        self, data_id: str = None, opener_id: str = None
    ) -> JsonObjectSchema:
        self._assert_valid_opener_id(opener_id)
        if data_id is not None:
            self._assert_has_data(data_id)
            product_handler = ProductHandler.guess(data_id)
            return product_handler.get_open_data_params_schema()
        else:
            return JsonObjectSchema(
                title="Opening parameters for all supported Sentinel products.",
                properties={
                    key: ph.get_open_data_params_schema()
                    for (key, ph) in zip(
                        ProductHandler.registry.keys(), ProductHandler.registry.values()
                    )
                },
            )

    def open_data(
        self,
        data_id: str,
        opener_id: str = None,
        data_type: DataTypeLike = None,
        **open_params,
    ) -> xr.Dataset | MultiLevelDataset:
        self._assert_has_data(data_id)
        self._assert_valid_data_type(data_type)
        self._assert_valid_opener_id(opener_id)
        schema = self.get_open_data_params_schema()
        schema.validate_instance(open_params)

        # search for items
        bbox_wgs84 = reproject_bbox(
            open_params["bbox"], open_params["crs"], "EPSG:4326"
        )

        search_params = dict(
            collections=[data_id],
            datetime=open_params["time_range"],
            intersects=bbox_to_geojson(bbox_wgs84),
            query=open_params.get("query"),
        )
        catalog = pystac_client.Client.open(STAC_URL)
        items = list(catalog.search(**search_params).items())
        # filter deprecated items
        items = filter_items_deprecated(items)
        # fiter items with incorrectly assigned footprint
        items = filter_items_wrong_footprint(items)
        if len(items) == 0:
            raise DataStoreError(f"No items found for search_params {search_params}.")

        product_handler = ProductHandler.guess(data_id)
        return product_handler.open_data(items, **open_params)

    def describe_data(
        self, data_id: str, data_type: DataTypeLike = None
    ) -> DatasetDescriptor:
        self._assert_has_data(data_id)
        self._assert_valid_data_type(data_type)
        raise NotImplementedError("`describe_data` is not implemented, yet.")

    def search_data(
        self, data_type: DataTypeLike = None, **search_params
    ) -> Iterator[DatasetDescriptor]:
        raise NotImplementedError(
            "Search is not supported. Only Sentinel-2 L1C and L2A products "
            "are currently handled."
        )

    def get_search_params_schema(
        self, data_type: DataTypeLike = None
    ) -> JsonObjectSchema:
        self._assert_valid_data_type(data_type)
        return JsonObjectSchema(
            properties=dict(),
            required=[],
            additional_properties=True,
        )

    # Auxiliary internal functions
    def _assert_has_data(self, data_id: str) -> None:
        if not self.has_data(data_id):
            raise DataStoreError(f"Data resource {data_id!r} is not available.")

    @staticmethod
    def _assert_valid_opener_id(opener_id: str) -> None:
        if opener_id is not None and opener_id is not EOPF_ZARR_OPENR_ID:
            raise DataStoreError(
                f"Data opener identifier must be {EOPF_ZARR_OPENR_ID!r}, "
                f"but got {opener_id!r}."
            )

    def _assert_valid_data_type(self, data_type: DataTypeLike) -> None:
        if not self._is_valid_data_type(data_type):
            raise DataStoreError(
                f"Data type must be {DATASET_TYPE.alias!r} "
                f"or None, but got {data_type!r}."
            )

    @staticmethod
    def _is_valid_data_type(data_type: DataTypeLike) -> bool:
        return data_type is None or DATASET_TYPE.is_super_type_of(data_type)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from abc import ABC

import numpy as np
import pystac
import xarray as xr
from xcube.util.jsonschema import JsonObjectSchema
from xcube_resampling import rectify_dataset

from xcube_eopf.constants import (
    SCHEMA_ADDITIONAL_QUERY,
    SCHEMA_AGG_METHODS,
    SCHEMA_BBOX,
    SCHEMA_CRS,
    SCHEMA_INTERP_METHODS,
    SCHEMA_SPATIAL_RES,
    SCHEMA_TILE_SIZE,
    SCHEMA_TIME_RANGE,
    SCHEMA_VARIABLES,
)
from xcube_eopf.prodhandler import ProductHandler, ProductHandlerRegistry
from xcube_eopf.utils import (
    add_attributes,
    add_nominal_datetime,
    get_gridmapping,
    mosaic_spatial_take_first,
)

_TILE_SIZE = 1024  # native chunk size of EOPF Sen3 Zarr samples


class Sen3ProductHandler(ProductHandler, ABC):
    """General Sentinel-3 product handles, defining methods applicable
    for Ol1Err, Ol1Efr, Ol2Lfr, and Sl2Lst product.
    """

    def get_open_data_params_schema(self) -> JsonObjectSchema:
        return JsonObjectSchema(
            title="Opening parameters for Sentinel-2 products.",
            properties=dict(
                variables=SCHEMA_VARIABLES,
                spatial_res=SCHEMA_SPATIAL_RES,
                time_range=SCHEMA_TIME_RANGE,
                bbox=SCHEMA_BBOX,
                crs=SCHEMA_CRS,
                tile_size=SCHEMA_TILE_SIZE,
                query=SCHEMA_ADDITIONAL_QUERY,
                agg_methods=SCHEMA_AGG_METHODS,
                interp_methods=SCHEMA_INTERP_METHODS,
            ),
            required=["time_range", "bbox", "crs", "spatial_res"],
            additional_properties=False,
        )

    def open_data(self, items: list[pystac.Item], **open_params) -> xr.Dataset:
        # get STAC items grouped by solar day
        grouped_items = group_items(items)

        # generate cube by mosaicking and stacking tiles
        ds = generate_cube(grouped_items, **open_params)

        # add attributes
        ds = add_attributes(ds, grouped_items, **open_params)

        return ds


class Sen3Ol1EfrProductHandler(Sen3ProductHandler):
    data_id = "sentinel-3-olci-l1-efr"


# this will be added later, when the footprints are corrected. Note, this
# product spanns over half of the globe, and thus crosses the antimeridian in many
# datasets, resulting in a lot of falsely assigned STAC item's bbox and geometry.
# class Sen3Ol1ErrProductHandler(Sen3ProductHandler):
#     data_id = "sentinel-3-olci-l1-err"


class Sen3Ol2LfrProductHandler(Sen3ProductHandler):
    data_id = "sentinel-3-olci-l2-lfr"


# Broken data in: https://stac.browser.user.eopf.eodc.eu/collections/sentinel-3-olci-l2-lrr?.language=en
# class Sen3Ol2LrrProductHandler(Sen3ProductHandler):
#     data_id = "sentinel-3-olci-l2-lrr"


# complex data tree groups, implementation postponed;
# class Sen3Sl1RbtProductHandler(Sen3ProductHandler):
#     data_id = "sentinel-3-slstr-l1-rbt"


class Sen3Sl2LstProductHandler(Sen3ProductHandler):
    data_id = "sentinel-3-slstr-l2-lst"


def register(registry: ProductHandlerRegistry):
    registry.register(Sen3Ol1EfrProductHandler)
    # registry.register(Sen3Ol1ErrProductHandler)
    registry.register(Sen3Ol2LfrProductHandler)
    # registry.register(Sen3Ol2LrrProductHandler)
    # registry.register(Sen3Sl1RbtProductHandler)
    registry.register(Sen3Sl2LstProductHandler)


def group_items(items: list[pystac.Item]) -> xr.DataArray:
    items = add_nominal_datetime(items)

    # get dates and tile IDs of the items
    dates = []
    for item in items:
        dates.append(item.properties["datetime_nominal"].date())
    dates = np.unique(dates)

    # sort items by date and tile ID into a data array
    grouped_items = np.full(len(dates), None, dtype=object)
    for idx, item in enumerate(items):
        date = item.properties["datetime_nominal"].date()
        idx_date = np.where(dates == date)[0][0]
        if grouped_items[idx_date] is None:
            grouped_items[idx_date] = [item]
        else:
            grouped_items[idx_date].append(item)

    grouped_items = xr.DataArray(grouped_items, dims=("time",), coords=dict(time=dates))

    # replace date by datetime from first item
    dts = []
    for date in grouped_items.time.values:
        item = np.sum(grouped_items.sel(time=date).values)[0]
        dts.append(item.datetime.replace(tzinfo=None))
    grouped_items = grouped_items.assign_coords(
        time=np.array(dts, dtype="datetime64[ns]")
    )
    grouped_items["time"].encoding["units"] = "seconds since 1970-01-01"
    grouped_items["time"].encoding["calendar"] = "standard"

    return grouped_items


def generate_cube(grouped_items: xr.DataArray, **open_params) -> xr.Dataset:
    target_gm = get_gridmapping(
        open_params["bbox"],
        open_params["spatial_res"],
        open_params["crs"],
        open_params.get("tile_size", _TILE_SIZE),
    )
    dss_time = []
    for dt_idx, dt in enumerate(grouped_items.time.values):
        items = grouped_items.sel(time=dt).item()
        dss_spatial = []
        for item in items:
            ds = xr.open_dataset(
                item.assets["product"].href + "/measurements",
                engine="eopf-zarr",
                chunks={},
                **dict(op_mode="native"),
            )
            variables = open_params.get("variables")
            if variables is not None:
                ds = ds[variables]
            if "time_stamp" in ds.coords:
                ds = ds.drop_vars("time_stamp")
            ds["latitude"] = ds["latitude"].persist()
            ds["longitude"] = ds["longitude"].persist()
            ds = rectify_dataset(
                ds,
                target_gm=target_gm,
                interp_methods=open_params.get("interp_methods"),
                agg_methods=open_params.get("agg_methods"),
            )
            dss_spatial.append(ds)
        dss_time.append(mosaic_spatial_take_first(dss_spatial))
    ds_final = xr.concat(dss_time, dim="time", join="exact")
    ds_final = ds_final.assign_coords(dict(time=grouped_items.time))
    return ds_final



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from unittest import TestCase

import xarray as xr
from xcube.core.store import new_data_store

from xcube_eopf.utils import reproject_bbox

from .helper import timeit

allowed_open_time = 30  # seconds
show_chunking = False


class Sentinel2Test(TestCase):
    def setUp(self):
        self.store = new_data_store("eopf-zarr")
        self.bbox = [9.7, 53.4, 10.3, 53.7]
        self.crs_utm = "EPSG:32632"
        self.bbox_utm = reproject_bbox(self.bbox, "EPSG:4326", self.crs_utm)

    def test_open_data_sen2l1c(self):
        with timeit("open 'sentinel-2-l1c'") as result:
            ds = self.store.open_data(
                data_id="sentinel-2-l1c",
                bbox=self.bbox_utm,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=10,
                crs=self.crs_utm,
                variables=["b02", "b03", "b04"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04"], list(ds.data_vars))
        self.assertCountEqual(
            [2, 3394, 4023], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 1830, 1830],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    def test_open_data_sen2l2a(self):
        with timeit("open 'sentinel-2-l2a'") as result:
            ds = self.store.open_data(
                data_id="sentinel-2-l2a",
                bbox=self.bbox_utm,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=10,
                crs=self.crs_utm,
                variables=["b02", "b03", "b04", "scl"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04", "scl"], list(ds.data_vars))
        self.assertCountEqual(
            [3, 3394, 4023], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 1830, 1830],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    def test_open_data_sen2l2a_with_reprojection(self):
        with timeit("open 'sentinel-2-l2a'") as result:
            ds = self.store.open_data(
                data_id="sentinel-2-l2a",
                bbox=self.bbox,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=10 / 111320,  # meter in degree,
                crs="EPSG:4326",
                variables=["b02", "b03", "b04", "scl"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04", "scl"], list(ds.data_vars))
        self.assertCountEqual(
            [3, 6681, 3341], [ds.sizes["time"], ds.sizes["lon"], ds.sizes["lat"]]
        )
        self.assertEqual(
            [1, 1830, 1830],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lon"][0],
                ds.chunksizes["lat"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.


def register_product_handlers():
    from xcube_eopf.prodhandler import ProductHandler

    from .sentinel1 import register as register_s1
    from .sentinel2 import register as register_s2
    from .sentinel3 import register as register_s3

    register_s1(ProductHandler.registry)
    register_s2(ProductHandler.registry)
    register_s3(ProductHandler.registry)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from unittest import TestCase

import xarray as xr
from xcube.core.store import new_data_store

from xcube_eopf.utils import reproject_bbox

from .helper import timeit

allowed_open_time = 120  # seconds
show_chunking = False


class Sentinel3Test(TestCase):
    def setUp(self):
        self.store = new_data_store("eopf-zarr")
        self.bbox = [9.7, 53.4, 10.3, 53.7]
        self.crs = "EPSG:4326"
        self.crs_utm = "EPSG:32632"
        self.bbox_utm = reproject_bbox(self.bbox, self.crs, self.crs_utm)

    def test_open_data_sen3ol1efr(self):
        with timeit("open 'sentinel-3-olci-l1-efr'") as result:
            ds = self.store.open_data(
                data_id="sentinel-3-olci-l1-efr",
                bbox=self.bbox,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=300 / 111320,
                crs=self.crs,
                variables=["oa01_radiance", "oa02_radiance"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["oa01_radiance", "oa02_radiance"], list(ds.data_vars))
        self.assertCountEqual(
            [1, 113, 224], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
        )
        self.assertEqual(
            [1, 113, 224],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lat"][0],
                ds.chunksizes["lon"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    def test_open_data_sen3ol1efr_with_projection(self):
        with timeit("open 'sentinel-3-olci-l1-efr'") as result:
            ds = self.store.open_data(
                data_id="sentinel-3-olci-l1-efr",
                bbox=self.bbox_utm,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=300,
                crs=self.crs_utm,
                variables=["oa01_radiance", "oa02_radiance"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["oa01_radiance", "oa02_radiance"], list(ds.data_vars))
        self.assertCountEqual(
            [1, 115, 136], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 115, 136],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    # leave out due to wrong footprint in the STAC item
    # def test_open_data_ol1err(self):
    #     with timeit("open 'sentinel-3-olci-l1-err'") as result:
    #         ds = self.store.open_data(
    #             data_id="sentinel-3-olci-l1-err",
    #             bbox=self.bbox,
    #             time_range=["2025-05-01", "2025-05-07"],
    #             spatial_res=300 / 111320,
    #             crs=self.crs,
    #             variables=["oa01_radiance", "oa02_radiance"],
    #         )
    #     self.assertTrue(result.time_delta < allowed_open_time)
    #     self.assertIsInstance(ds, xr.Dataset)
    #     self.assertCountEqual(["oa01_radiance", "oa02_radiance"], list(ds.data_vars))
    #     self.assertCountEqual(
    #         [1, 113, 224], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
    #     )
    #     self.assertEqual(
    #         [1, 113, 224],
    #         [
    #             ds.chunksizes["time"][0],
    #             ds.chunksizes["lat"][0],
    #             ds.chunksizes["lon"][0],
    #         ],
    #     )
    #     self.assertIn("stac_url", ds.attrs)
    #     self.assertIn("stac_items", ds.attrs)
    #     self.assertIn("open_params", ds.attrs)
    #     self.assertIn("xcube_eopf_version", ds.attrs)

    def test_open_data_ol2lfr(self):
        with timeit("open 'sentinel-3-olci-l2-lfr'") as result:
            ds = self.store.open_data(
                data_id="sentinel-3-olci-l2-lfr",
                bbox=self.bbox,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=300 / 111320,  # meter in degree,
                crs="EPSG:4326",
                variables=["gifapar", "iwv"],
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["gifapar", "iwv"], list(ds.data_vars))
        self.assertCountEqual(
            [1, 113, 224], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
        )
        self.assertEqual(
            [1, 113, 224],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lat"][0],
                ds.chunksizes["lon"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    def test_open_data_sl2lst(self):
        with timeit("open 'sentinel-3-slstr-l2-lst'") as result:
            ds = self.store.open_data(
                data_id="sentinel-3-slstr-l2-lst",
                bbox=self.bbox,
                time_range=["2025-05-01", "2025-05-07"],
                spatial_res=300 / 111320,  # meter in degree,
                crs="EPSG:4326",
            )
        self.assertTrue(result.time_delta < allowed_open_time)
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["lst"], list(ds.data_vars))
        self.assertCountEqual(
            [1, 113, 224], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
        )
        self.assertEqual(
            [1, 113, 224],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lat"][0],
                ds.chunksizes["lon"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)



import time


class timeit:
    """A context manager used to measure time it takes
    to execute its with-block.
    The result is available as `time_delta` attribute.

    Args:
        label: A text label
        silent: Whether to suppress printing the result
    """

    def __init__(self, label: str | None = None, silent: bool = False):
        self.label = label
        self.silent = silent
        self.start_time: float | None = None
        self.time_delta: float | None = None

    def __enter__(self) -> "timeit":
        self.start_time = time.process_time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.time_delta = time.process_time() - self.start_time
        if not self.silent:
            print(f"{self.label or 'code block'} took {self.time_delta:.3f} seconds")



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import unittest

from xcube.util.extension import ExtensionRegistry

from xcube_eopf.plugin import init_plugin


class XcubePluginTest(unittest.TestCase):
    def test_plugin(self):
        """Assert xcube extensions registered by xcube-stac"""
        registry = ExtensionRegistry()
        init_plugin(registry)
        self.assertEqual(
            {
                "xcube.core.store": {
                    "eopf-zarr": {
                        "component": "<not loaded yet>",
                        "description": "EOPF-Zarr DataStore",
                        "name": "eopf-zarr",
                        "point": "xcube.core.store",
                    },
                }
            },
            registry.to_dict(),
        )



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import datetime
import unittest

import numpy as np
import pyproj
import pystac
import xarray as xr
from xcube.core.store import DataStoreError
from xcube_resampling.gridmapping import GridMapping

from xcube_eopf.utils import (
    add_nominal_datetime,
    get_gridmapping,
    get_spatial_dims,
    mosaic_spatial_take_first,
    normalize_crs,
    reproject_bbox,
)


class UtilsTest(unittest.TestCase):

    def test_reproject_bbox(self):
        bbox_wgs84 = [2, 50, 3, 51]
        crs_wgs84 = "EPSG:4326"
        crs_3035 = "EPSG:3035"
        bbox_3035 = [3748675.9529771, 3011432.8944597, 3830472.1359979, 3129432.4914285]
        self.assertEqual(bbox_wgs84, reproject_bbox(bbox_wgs84, crs_wgs84, crs_wgs84))
        self.assertEqual(bbox_3035, reproject_bbox(bbox_3035, crs_3035, crs_3035))
        np.testing.assert_almost_equal(
            reproject_bbox(bbox_wgs84, crs_wgs84, crs_3035), bbox_3035
        )
        np.testing.assert_almost_equal(
            reproject_bbox(
                reproject_bbox(bbox_wgs84, crs_wgs84, crs_3035, buffer=0.0),
                crs_3035,
                crs_wgs84,
                buffer=0.0,
            ),
            [
                1.829619451017442,
                49.93464594063249,
                3.1462425554926226,
                51.06428203128216,
            ],
        )

        crs_utm = "EPSG:32601"
        bbox_utm = [
            213372.0489639729,
            5540547.369934658,
            362705.63410562894,
            5768595.563692021,
        ]
        np.testing.assert_almost_equal(
            reproject_bbox(bbox_utm, crs_utm, crs_wgs84, buffer=0.02),
            [178.77930769, 49.90632759, -178.87064939, 52.09298731],
        )

    def test_normalize_crs(self):
        crs_str = "EPSG:4326"
        crs_pyproj = pyproj.CRS.from_string(crs_str)
        self.assertEqual(crs_pyproj, normalize_crs(crs_str))
        self.assertEqual(crs_pyproj, normalize_crs(crs_pyproj))

    def test_add_nominal_datetime(self):
        item0 = pystac.Item(
            id="item0",
            geometry=None,
            bbox=[0, 50, 1, 51],
            datetime=datetime.datetime(2024, 6, 1, 12, 0, 0),
            properties=dict(),
        )
        item1 = pystac.Item(
            id="item1",
            geometry=None,
            bbox=[150, 50, 151, 51],
            datetime=datetime.datetime(2024, 7, 1, 16, 0, 0),
            properties=dict(),
        )
        items = [item0, item1]
        items_nominal_time = add_nominal_datetime(items)

        item = items_nominal_time[0]
        self.assertIn("datetime_nominal", item.properties)
        self.assertEqual(
            datetime.datetime(2024, 6, 1, 12, 0, 0),
            item.properties["datetime_nominal"],
        )
        self.assertIn("center_point", item.properties)
        self.assertEqual((0.5, 50.5), item.properties["center_point"])

        item = items_nominal_time[1]
        self.assertIn("datetime_nominal", item.properties)
        self.assertEqual(
            datetime.datetime(2024, 7, 2, 2, 0, 0),
            item.properties["datetime_nominal"],
        )
        self.assertIn("center_point", item.properties)
        self.assertEqual((150.5, 50.5), item.properties["center_point"])

    def test_mosaic_spatial_take_first(self):
        list_ds = []
        # first tile
        data = np.array(
            [
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                [[10, 11, 12], [13, 14, np.nan], [np.nan, np.nan, np.nan]],
                [[19, 20, 21], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]],
            ],
            dtype=float,
        )
        dims = ("time", "lat", "lon")
        coords = {
            "time": np.array(
                ["2025-01-01", "2025-01-02", "2025-01-03"], dtype="datetime64"
            ),
            "lat": [10.0, 20.0, 30.0],
            "lon": [100.0, 110.0, 120.0],
        }
        da = xr.DataArray(data, dims=dims, coords=coords)
        list_ds.append(xr.Dataset({"B01": da}))
        # second tile
        data = np.array(
            [
                [[np.nan, np.nan, np.nan], [np.nan, np.nan, 106], [107, 108, 109]],
                [[np.nan, np.nan, np.nan], [113, 114, 115], [116, 117, 118]],
                [[np.nan, np.nan, 120], [121, 122, 123], [124, 125, 126]],
            ],
            dtype=float,
        )
        dims = ("time", "lat", "lon")
        coords = {
            "time": np.array(
                ["2025-01-01", "2025-01-02", "2025-01-03"], dtype="datetime64"
            ),
            "lat": [10.0, 20.0, 30.0],
            "lon": [100.0, 110.0, 120.0],
        }
        da = xr.DataArray(data, dims=dims, coords=coords)
        list_ds.append(xr.Dataset({"B01": da}))

        # test only one tile
        ds_test = mosaic_spatial_take_first(list_ds[:1])
        self.assertIsInstance(ds_test, xr.Dataset)
        xr.testing.assert_allclose(ds_test, list_ds[0])

        # test two tiles
        ds_test = mosaic_spatial_take_first(list_ds)
        self.assertIsInstance(ds_test, xr.Dataset)
        data = np.array(
            [
                [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                [[10, 11, 12], [13, 14, 115], [116, 117, 118]],
                [[19, 20, 21], [121, 122, 123], [124, 125, 126]],
            ],
            dtype=float,
        )
        dims = ("time", "lat", "lon")
        coords = {
            "time": np.array(
                ["2025-01-01", "2025-01-02", "2025-01-03"],
                dtype="datetime64",
            ),
            "lat": [10.0, 20.0, 30.0],
            "lon": [100.0, 110.0, 120.0],
        }
        da = xr.DataArray(data, dims=dims, coords=coords)
        ds_expected = xr.Dataset({"B01": da})
        xr.testing.assert_allclose(ds_test, ds_expected)

        # test two tiles, where spatial ref is given in spatial_ref coord
        spatial_ref = xr.DataArray(np.array(0), attrs=dict(crs_wkt="testing"))
        for i, ds in enumerate(list_ds):
            ds.coords["spatial_ref"] = spatial_ref
            list_ds[i] = ds
        ds_expected = xr.Dataset({"B01": da})
        ds_expected = ds_expected.assign_coords({"spatial_ref": spatial_ref})
        ds_test = mosaic_spatial_take_first(list_ds)
        self.assertIsInstance(ds_test, xr.Dataset)
        xr.testing.assert_allclose(ds_test, ds_expected)

    def test_get_spatial_dims(self):
        ds = xr.Dataset()
        ds["var"] = xr.DataArray(
            data=np.ones((2, 2)), dims=("y", "x"), coords=dict(y=[0, 10], x=[0, 10])
        )
        self.assertEqual(("y", "x"), get_spatial_dims(ds))
        ds = xr.Dataset()
        ds["var"] = xr.DataArray(
            data=np.ones((2, 2)),
            dims=("lat", "lon"),
            coords=dict(lat=[0, 10], lon=[0, 10]),
        )
        self.assertEqual(("lat", "lon"), get_spatial_dims(ds))
        ds = xr.Dataset()
        ds["var"] = xr.DataArray(
            data=np.ones((2, 2)),
            dims=("dim_false0", "dim_false1"),
            coords=dict(dim_false0=[0, 10], dim_false1=[0, 10]),
        )
        with self.assertRaises(DataStoreError) as cm:
            get_spatial_dims(ds)
        self.assertEqual(
            "No spatial dimensions found in dataset.",
            f"{cm.exception}",
        )

    def test_get_gridmapping(self):
        bbox = [0, 50, 1, 51]
        spatial_res = 0.0001
        crs = "EPSG:4326"
        tile_size = 2000
        gm = get_gridmapping(bbox, spatial_res, crs, tile_size)
        self.assertIsInstance(gm, GridMapping)
        self.assertEqual((0.0001, 0.0001), gm.xy_res)
        self.assertEqual(pyproj.CRS.from_epsg(4326), gm.crs)
        self.assertEqual((2000, 2000), gm.tile_size)
        self.assertEqual((10001, 10001), gm.size)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import logging

from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonComplexSchema,
    JsonDateSchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObjectSchema,
    JsonStringSchema,
)
from xcube_resampling.constants import AGG_METHODS

# general stac constants
DATA_STORE_ID = "eopf-zarr"
STAC_URL = "https://stac.core.eopf.eodc.eu"
EOPF_ZARR_OPENR_ID = "dataset:zarr:eopf-zarr"

# other constants
CONVERSION_FACTOR_DEG_METER = 111320
LOG = logging.getLogger("xcube.eopf")

# general schema definition
SCHEMA_ADDITIONAL_QUERY = JsonObjectSchema(
    additional_properties=True,
    title="Additional query options used during item search of STAC API.",
    description=(
        "Additional filtering based on the properties of Item objects "
        "is supported. For more information see "
        "https://github.com/stac-api-extensions/query"
    ),
)
SCHEMA_BBOX = JsonArraySchema(
    items=(
        JsonNumberSchema(),
        JsonNumberSchema(),
        JsonNumberSchema(),
        JsonNumberSchema(),
    ),
    title="Bounding box [west, south, east, north] in geographical coordinates.",
)
SCHEMA_TIME_RANGE = JsonArraySchema(
    items=[
        JsonDateSchema(nullable=True),
        JsonDateSchema(nullable=True),
    ],
    title="Time Range",
    description=(
        "Time range given as pair of start and stop dates. "
        "Dates must be given using format 'YYYY-MM-DD'. "
        "Start and stop are inclusive."
    ),
)
SCHEMA_VARIABLES = JsonComplexSchema(
    title="Names of variables in dataset",
    description="Names of variables which will be included in the data cube.",
    one_of=[
        JsonStringSchema(
            title="Variable name or regex pattern",
            min_length=0,
        ),
        JsonArraySchema(
            title="Iterable of variables",
            items=(JsonStringSchema(min_length=0)),
            unique_items=True,
        ),
    ],
)
SCHEMA_SPATIAL_RES = JsonNumberSchema(title="Spatial Resolution", exclusive_minimum=0.0)
SCHEMA_CRS = JsonStringSchema(title="Coordinate reference system")
SCHEMA_TILE_SIZE = JsonArraySchema(
    nullable=True,
    title="Tile size of returned dataset",
    description=(
        "Spatial tile size in y and x (or lat and lon if crs is geographic) "
        "in returned dataset."
    ),
    items=[JsonIntegerSchema(minimum=1), JsonIntegerSchema(minimum=1)],
)
SCHEMA_INTERP_METHODS = JsonComplexSchema(
    title="Interpolation method for updampling",
    description=(
        "Specifies the interpolation method used for upsampling spatial data variables. "
        "You can provide a single method for all variables or a dictionary mapping "
        "variable names or data types to specific interpolation methods. "
        "For detailed documentation, see https://xcube-dev.github.io/xcube-resampling/."
    ),
    one_of=[
        JsonIntegerSchema(enum=[0, 1], description="0: nearest neighbor; 1: bilinear"),
        JsonStringSchema(enum=["nearest", "triangular", "bilinear"]),
        JsonObjectSchema(
            title=(
                "dictionary mapping variable names or data types"
                " to specific interpolation methods."
            )
        ),
    ],
)

SCHEMA_AGG_METHODS = JsonComplexSchema(
    title="Aggregation methods for downsampling",
    description=(
        "Specifies the aggregation method used for downsampling spatial data variables. "
        "You can provide a single method for all variables or a dictionary mapping "
        "variable names or data types to specific aggregation methods. "
        "For detailed documentation, see https://xcube-dev.github.io/xcube-resampling/."
    ),
    one_of=[
        JsonStringSchema(enum=list(AGG_METHODS.keys())),
        JsonObjectSchema(
            title=(
                "dictionary mapping variable names or data types"
                " to specific aggregation methods."
            )
        ),
    ],
)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import datetime
from unittest import TestCase

import pystac
import xarray as xr

from xcube_eopf.prodhandlers.sentinel3 import group_items


class Sentinel3Test(TestCase):

    def test_group_items(self):
        item0 = pystac.Item(
            id="S2A_MSIL2A_20240604T103031_N0500_R108_T32UQD_20240604T120000",
            geometry=None,
            bbox=[0, 50, 1, 51],
            datetime=datetime.datetime(2024, 6, 4, 10, 30, 31),
            properties={
                "datetime_nominal": datetime.datetime(2024, 6, 4, 10, 30, 31),
            },
        )
        item1 = pystac.Item(
            id="S2A_MSIL2A_20240604T103031_N0500_R108_T32UQD_20240604T120000",
            geometry=None,
            bbox=[0, 50, 1, 51],
            datetime=datetime.datetime(2024, 6, 4, 10, 30, 31),
            properties={
                "datetime_nominal": datetime.datetime(2024, 6, 4, 10, 30, 31),
            },
        )

        grouped_item = group_items([item0, item1])
        self.assertIsInstance(grouped_item, xr.DataArray)
        self.assertEqual(dict(time=1), grouped_item.sizes)
        self.assertIsInstance(grouped_item[0].item(), list)
        self.assertEqual(2, len(grouped_item[0].item()))



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from xcube_eopf.prodhandler import ProductHandlerRegistry


def register(_registry: ProductHandlerRegistry):
    # TODO: implement Sentinel-1 ProductHandler
    pass



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from unittest import TestCase
from unittest.mock import patch

import numpy as np
import pytest
import xarray as xr
from xcube.core.store import DataStoreError, new_data_store
from xcube.util.jsonschema import JsonObjectSchema

from xcube_eopf.constants import CONVERSION_FACTOR_DEG_METER, DATA_STORE_ID
from xcube_eopf.utils import reproject_bbox

from .helpers import sen2_l2a_10m, sen2_l2a_60m, sen2_l2a_60m_wo_scl, sen3_ol1efr_data


class EOPFZarrDataStoreTest(TestCase):

    def setUp(self):
        self.store = new_data_store(DATA_STORE_ID)

    def test_get_data_store_params_schema(self):
        schema = self.store.get_data_store_params_schema()
        self.assertIsInstance(schema, JsonObjectSchema)
        self.assertEqual(dict(), schema.properties)
        self.assertEqual([], schema.required)
        self.assertFalse(schema.additional_properties)

    def test_get_data_types(self):
        self.assertEqual(("dataset",), self.store.get_data_types())

    def test_get_data_types_for_data(self):
        self.assertEqual(
            ("dataset",), self.store.get_data_types_for_data("sentinel-2-l1c")
        )
        self.assertEqual(
            ("dataset",), self.store.get_data_types_for_data("sentinel-2-l2a")
        )

    def test_get_data_ids(self):
        self.assertCountEqual(
            [
                "sentinel-2-l1c",
                "sentinel-2-l2a",
                "sentinel-3-olci-l2-lfr",
                "sentinel-3-olci-l1-efr",
                "sentinel-3-slstr-l2-lst",
            ],
            self.store.get_data_ids(),
        )
        self.assertCountEqual(
            [
                ("sentinel-2-l1c", {}),
                ("sentinel-2-l2a", {}),
                ("sentinel-3-olci-l2-lfr", {}),
                ("sentinel-3-olci-l1-efr", {}),
                ("sentinel-3-slstr-l2-lst", {}),
            ],
            self.store.get_data_ids(include_attrs=True),
        )

    def test_has_data(self):
        self.assertTrue(self.store.has_data("sentinel-2-l1c"))
        self.assertTrue(self.store.has_data("sentinel-2-l2a"))
        self.assertTrue(self.store.has_data("sentinel-3-olci-l2-lfr"))
        self.assertFalse(self.store.has_data("sentinel-2-l3a"))
        self.assertTrue(self.store.has_data("sentinel-2-l2a", data_type="dataset"))
        with self.assertRaises(DataStoreError) as cm:
            self.store.has_data("sentinel-2-l2a", data_type="mldataset")
        self.assertEqual(
            "Data type must be 'dataset' or None, but got 'mldataset'.",
            str(cm.exception),
        )

    def test_get_data_opener_ids(self):
        self.assertEqual(("dataset:zarr:eopf-zarr",), self.store.get_data_opener_ids())
        self.assertEqual(
            ("dataset:zarr:eopf-zarr",),
            self.store.get_data_opener_ids(data_id="sentinel-2-l2a"),
        )
        with self.assertRaises(DataStoreError) as cm:
            self.store.get_data_opener_ids("sentinel-2-l3a")
        self.assertEqual(
            "Data resource 'sentinel-2-l3a' is not available.",
            str(cm.exception),
        )

    def test_get_open_data_params_schema(self):
        # no optional arguments; get all open parameters from all product handlers
        schema = self.store.get_open_data_params_schema()
        self.assertIsInstance(schema, JsonObjectSchema)
        self.assertIn("sentinel-2-l1c", schema.properties)
        self.assertIn("sentinel-2-l2a", schema.properties)
        self.assertIn("sentinel-3-olci-l2-lfr", schema.properties)
        self.assertIn("sentinel-3-olci-l1-efr", schema.properties)
        self.assertIn("sentinel-3-slstr-l2-lst", schema.properties)

        # with data_id argument for Sen2
        schema = self.store.get_open_data_params_schema(data_id="sentinel-2-l2a")
        self.assertIsInstance(schema, JsonObjectSchema)
        self.assertIn("variables", schema.properties)
        self.assertIn("spatial_res", schema.properties)
        self.assertIn("time_range", schema.properties)
        self.assertIn("bbox", schema.properties)
        self.assertIn("crs", schema.properties)
        self.assertIn("tile_size", schema.properties)
        self.assertIn("query", schema.properties)
        self.assertNotIn("test", schema.properties)
        self.assertCountEqual(
            ["time_range", "bbox", "crs", "spatial_res"], schema.required
        )

        # with data_id argument for Sen3
        schema = self.store.get_open_data_params_schema(
            data_id="sentinel-3-olci-l1-efr"
        )
        self.assertIsInstance(schema, JsonObjectSchema)
        self.assertIn("variables", schema.properties)
        self.assertIn("spatial_res", schema.properties)
        self.assertIn("time_range", schema.properties)
        self.assertIn("bbox", schema.properties)
        self.assertIn("crs", schema.properties)
        self.assertIn("tile_size", schema.properties)
        self.assertIn("query", schema.properties)
        self.assertNotIn("test", schema.properties)
        self.assertCountEqual(
            ["time_range", "bbox", "crs", "spatial_res"], schema.required
        )

        # with invalid opener_id
        with self.assertRaises(DataStoreError) as cm:
            self.store.get_open_data_params_schema(opener_id="dataset:netcdf:eopf-zarr")
        self.assertEqual(
            "Data opener identifier must be 'dataset:zarr:eopf-zarr', "
            "but got 'dataset:netcdf:eopf-zarr'.",
            str(cm.exception),
        )

    @pytest.mark.vcr()
    @patch("xarray.open_dataset")
    def test_open_data_sen2_10m(self, mock_xarray):
        mock_xarray.return_value = sen2_l2a_10m()

        ds = self.store.open_data(
            data_id="sentinel-2-l2a",
            bbox=(610000, 5880000, 630000, 5900000),
            time_range=["2025-05-01", "2025-05-15"],
            spatial_res=10,
            crs="EPSG:32632",
            variables=["b02", "b03", "b04", "scl"],
            interp_methods={np.float32: 0, "scl": 3},
        )
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04", "scl"], list(ds.data_vars))
        self.assertEqual(
            [4, 2000, 2000], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 1830, 1830],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    @pytest.mark.vcr()
    @patch("xarray.open_dataset")
    def test_open_data_sen2_5m_interp_method(self, mock_xarray):
        mock_xarray.return_value = sen2_l2a_10m()
        ds = self.store.open_data(
            data_id="sentinel-2-l2a",
            bbox=(610000, 5880000, 630000, 5900000),
            time_range=["2025-05-01", "2025-05-15"],
            spatial_res=5,
            crs="EPSG:32632",
            variables=["b02", "b03", "b04", "scl"],
            interp_methods={np.dtype("float64"): "bilinear", "scl": "nearest"},
        )
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04", "scl"], list(ds.data_vars))
        self.assertEqual(
            [4, 4001, 4001], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 1830, 1830],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    @pytest.mark.vcr()
    @patch("xarray.open_dataset")
    def test_open_data_sen2_100m_agg_methods(self, mock_xarray):
        mock_xarray.return_value = sen2_l2a_60m()
        ds = self.store.open_data(
            data_id="sentinel-2-l2a",
            bbox=(610000, 5880000, 630000, 5900000),
            time_range=["2025-05-01", "2025-05-15"],
            spatial_res=100,
            crs="EPSG:32632",
            variables=["b02", "b03", "b04", "scl"],
            agg_methods={np.dtype("float64"): "max", "scl": "mode"},
        )
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04", "scl"], list(ds.data_vars))
        self.assertEqual(
            [4, 201, 201], [ds.sizes["time"], ds.sizes["y"], ds.sizes["x"]]
        )
        self.assertEqual(
            [1, 201, 201],
            [ds.chunksizes["time"][0], ds.chunksizes["y"][0], ds.chunksizes["x"][0]],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    @pytest.mark.vcr()
    @patch("xarray.open_dataset")
    def test_open_data_sen2_geographic(self, mock_xarray):
        mock_xarray.return_value = sen2_l2a_60m_wo_scl()

        # open Sentinel-2 L2A
        bbox = [610000, 5880000, 630000, 5900000]
        bbox_wgs84 = reproject_bbox(bbox, "EPSG:32632", "EPSG:4326")
        ds = self.store.open_data(
            data_id="sentinel-2-l2a",
            bbox=bbox_wgs84,
            time_range=["2025-05-01", "2025-05-15"],
            spatial_res=50 / CONVERSION_FACTOR_DEG_METER,
            crs="EPSG:4326",
            variables=["b02", "b03", "b04"],
        )
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["b02", "b03", "b04"], list(ds.data_vars))
        self.assertEqual(
            [4, 412, 684], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
        )
        self.assertEqual(
            [1, 412, 684],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lat"][0],
                ds.chunksizes["lon"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    @pytest.mark.vcr()
    @patch("xarray.open_dataset")
    def test_open_data_sen3_geographic(self, mock_xarray):
        mock_xarray.return_value = sen3_ol1efr_data()

        # open Sentinel-3 OL1EFR
        bbox = [5.0, 53.0, 10.0, 57.0]
        ds = self.store.open_data(
            data_id="sentinel-3-olci-l1-efr",
            bbox=bbox,
            time_range=["2025-05-01", "2025-05-21"],
            spatial_res=300 / CONVERSION_FACTOR_DEG_METER,
            crs="EPSG:4326",
            variables=["oa01_radiance", "oa02_radiance"],
        )
        self.assertIsInstance(ds, xr.Dataset)
        self.assertCountEqual(["oa01_radiance", "oa02_radiance"], list(ds.data_vars))
        self.assertEqual(
            [2, 1486, 1857], [ds.sizes["time"], ds.sizes["lat"], ds.sizes["lon"]]
        )
        self.assertEqual(
            [1, 1024, 1024],
            [
                ds.chunksizes["time"][0],
                ds.chunksizes["lat"][0],
                ds.chunksizes["lon"][0],
            ],
        )
        self.assertIn("stac_url", ds.attrs)
        self.assertIn("stac_items", ds.attrs)
        self.assertIn("open_params", ds.attrs)
        self.assertIn("xcube_eopf_version", ds.attrs)

    @pytest.mark.vcr()
    def test_open_data_no_items_found(self):
        with self.assertRaises(DataStoreError) as cm:
            _ = self.store.open_data(
                data_id="sentinel-2-l2a",
                bbox=(610000, 5880000, 630000, 5900000),
                time_range=["2016-05-01", "2017-05-15"],
                spatial_res=10,
                crs="EPSG:32632",
                variables=["b02", "b03", "b04", "scl"],
            )
        self.assertIn("No items found for search_params", str(cm.exception))

    def test_describe_data(self):
        with self.assertRaises(NotImplementedError) as cm:
            self.store.describe_data("sentinel-2-l1c")
        self.assertEqual("`describe_data` is not implemented, yet.", f"{cm.exception}")

    def test_search_data(self):
        with self.assertRaises(NotImplementedError) as cm:
            self.store.search_data("sentinel-2-l1c")
        self.assertEqual(
            (
                "Search is not supported. Only Sentinel-2 L1C and L2A products "
                "are currently handled."
            ),
            f"{cm.exception}",
        )

    def test_get_search_params_schema(self):
        schema = self.store.get_search_params_schema()
        self.assertEqual(schema.properties, dict())
        self.assertEqual(schema.required, [])
        self.assertTrue(schema.additional_properties)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from .sentinel2 import sen2_l2a_10m, sen2_l2a_60m, sen2_l2a_60m_wo_scl
from .sentinel3 import sen3_ol1efr_data

__all__ = ["sen2_l2a_10m", "sen2_l2a_60m", "sen2_l2a_60m_wo_scl", "sen3_ol1efr_data"]



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from unittest import TestCase

import pytest
import xarray as xr
from xcube.util.jsonschema import JsonObjectSchema

from xcube_eopf.prodhandler import ProductHandler, ProductHandlerRegistry
from xcube_eopf.prodhandlers.sentinel2 import (
    Sen2L1CProductHandler,
    Sen2L2AProductHandler,
)


class TestProductHandler(ProductHandler):
    data_id = "TEST"

    def get_open_data_params_schema(self) -> JsonObjectSchema:
        return JsonObjectSchema()

    def open_data(self, **open_params) -> xr.Dataset:
        return xr.Dataset()


class ProductHandlerTest(TestCase):
    def setUp(self):
        ProductHandler.registry.register(TestProductHandler)

    def tearDown(self):
        ProductHandler.registry.unregister(TestProductHandler)

    def test_guess_ok(self):
        self.assertIsInstance(ProductHandler.guess("TEST"), TestProductHandler)

    # noinspection PyMethodMayBeStatic
    def test_guess_fail(self):
        with pytest.raises(
            ValueError,
            match=(
                "Unable to detect product handler for 'REST'. Use `data_id` "
                "argument to pass one of 'TEST'."
            ),
        ):
            ProductHandler.guess("REST")

    def test_from_data_id(self):
        self.assertIsInstance(ProductHandler.from_data_id("TEST"), TestProductHandler)
        self.assertIsNone(ProductHandler.from_data_id("REST"))


class ProductHandlerRegistryTest(TestCase):
    # noinspection PyMethodMayBeStatic
    def get(self):
        reg = ProductHandlerRegistry()
        reg.register(Sen2L1CProductHandler)
        reg.register(Sen2L2AProductHandler)
        return reg

    def test_get(self):
        reg = self.get()
        self.assertIsInstance(reg.get("sentinel-2-l1c"), Sen2L1CProductHandler)
        self.assertIsInstance(reg.get("sentinel-2-l2a"), Sen2L2AProductHandler)
        self.assertIsNone(reg.get("sentinel-2-l3a"))

    def test_keys_and_values(self):
        reg = self.get()
        self.assertEqual(["sentinel-2-l1c", "sentinel-2-l2a"], list(reg.keys()))
        values = list(reg.values())
        self.assertEqual(2, len(values))
        self.assertIsInstance(values[0], Sen2L1CProductHandler)
        self.assertIsInstance(values[1], Sen2L2AProductHandler)

    def test_register_unregister(self):
        reg = self.get()
        reg.register(TestProductHandler)
        self.assertIsInstance(reg.get("TEST"), TestProductHandler)
        reg.unregister(TestProductHandler)
        self.assertIsNone(reg.get("TEST"))



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import dask.array as da
import numpy as np
import xarray as xr


def sen3_ol1efr_data():
    height = 4091
    width = 4865
    tile_height = 1024
    tile_width = 1024

    # band data
    bands = [f"oa{i:02d}_radiance" for i in range(1, 22)]
    mock_data = {
        band: (
            ("rows", "columns"),
            da.ones(
                (height, width), chunks=(tile_height, tile_width), dtype=np.float32
            ),
        )
        for band in bands
    }

    # geolocation data
    lon = da.linspace(0, 15, width, chunks=tile_width, dtype=np.float32)
    lon *= da.linspace(0.5, 1.5, width, chunks=tile_width, dtype=np.float32)
    lat = da.linspace(50, 60, height, chunks=tile_height, dtype=np.float32)
    lat *= da.linspace(0.5, 1.5, height, chunks=tile_height, dtype=np.float32)
    lon, lat = da.meshgrid(lon, lat, indexing="xy")
    coords = {
        "longitude": (("rows", "columns"), lon),
        "latitude": (("rows", "columns"), lat),
        "time_stamp": (("rows",), np.arange(height).astype("datetime64[ns]")),
    }
    return xr.Dataset(mock_data, coords=coords)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from abc import ABC
from collections import defaultdict

import dask.array as da
import numpy as np
import pyproj
import pystac
import xarray as xr
from xcube.util.jsonschema import JsonObjectSchema
from xcube_resampling.gridmapping import GridMapping
from xcube_resampling.spatial import resample_in_space

from xcube_eopf.constants import (
    CONVERSION_FACTOR_DEG_METER,
    SCHEMA_ADDITIONAL_QUERY,
    SCHEMA_AGG_METHODS,
    SCHEMA_BBOX,
    SCHEMA_CRS,
    SCHEMA_INTERP_METHODS,
    SCHEMA_SPATIAL_RES,
    SCHEMA_TILE_SIZE,
    SCHEMA_TIME_RANGE,
    SCHEMA_VARIABLES,
)
from xcube_eopf.prodhandler import ProductHandler, ProductHandlerRegistry
from xcube_eopf.utils import (
    add_attributes,
    add_nominal_datetime,
    get_gridmapping,
    mosaic_spatial_take_first,
    normalize_crs,
    reproject_bbox,
)

_SEN2_SPATIAL_RES = np.array([10, 20, 60])
_TILE_SIZE = 1830  # chunk size of 10m resolution and a multiple of 20m and 60m
_ATTRIBUTE_KEYS = [
    "_eopf_attrs",
    "long_name",
    "flag_values",
    "flag_meanings",
    "flag_colors",
    "grid_mapping",
]


class Sen2ProductHandler(ProductHandler, ABC):
    """General Sentinel-2 product handles, defining methods applicable
    for L1C and L2A product.
    """

    # noinspection PyMethodMayBeStatic
    def get_open_data_params_schema(self) -> JsonObjectSchema:
        return JsonObjectSchema(
            title="Opening parameters for Sentinel-2 products.",
            properties=dict(
                variables=SCHEMA_VARIABLES,
                spatial_res=SCHEMA_SPATIAL_RES,
                time_range=SCHEMA_TIME_RANGE,
                bbox=SCHEMA_BBOX,
                crs=SCHEMA_CRS,
                tile_size=SCHEMA_TILE_SIZE,
                query=SCHEMA_ADDITIONAL_QUERY,
                agg_methods=SCHEMA_AGG_METHODS,
                interp_methods=SCHEMA_INTERP_METHODS,
            ),
            required=["time_range", "bbox", "crs", "spatial_res"],
            additional_properties=False,
        )

    def open_data(self, items: list[pystac.Item], **open_params) -> xr.Dataset:
        # get STAC items grouped by solar day and MRGS tile
        grouped_items = group_items(items)

        # generate cube by mosaicking and stacking tiles
        ds = generate_cube(grouped_items, **open_params)

        # add attributes
        ds = add_attributes(ds, grouped_items, **open_params)

        # TODO how to handle solar and viewing angles

        return ds


class Sen2L1CProductHandler(Sen2ProductHandler):
    data_id = "sentinel-2-l1c"


class Sen2L2AProductHandler(Sen2ProductHandler):
    data_id = "sentinel-2-l2a"


def register(registry: ProductHandlerRegistry):
    registry.register(Sen2L1CProductHandler)
    registry.register(Sen2L2AProductHandler)


def group_items(items: list[pystac.Item]) -> xr.DataArray:
    """Group STAC items by solar day and tile ID.

    Organizes a list of Sentinel-2 STAC items into an `xarray.DataArray` with
    dimensions `(time, tile_id)`, where:
    - `time` represents the solar acquisition date (date only, time of day is ignored).
    - `tile_id` corresponds to the Sentinel-2 MGRS tile code.

    Args:
        items: A list of STAC items. Each item must include:
            - `properties["datetime_nominal"]`: The nominal acquisition datetime.
            - `properties["grid:code"]`: The MGRS tile code.

    Returns:
        A 2D `xarray.DataArray` of shape (time, tile_id), where each cell contains a
        list of STAC items for the given date and tile ID. The `time` coordinate is
        derived from the nominal acquisition datetime of the first item in each
        (date, tile) group.

    Notes:
        - Each cell in the returned array contains a list of items because tiles may
          be split across multiple files.
    """
    items = add_nominal_datetime(items)

    # TODO: So far no handling of processing version,
    #       STAC items with multiple processing versions are added to the list and
    #       mosaicked by taking the fist non-NaN value.
    #       For proper handling wait for STAC item update (see https://github.com/EOPF-Sample-Service/eopf-stac/issues/28)
    # get dates and tile IDs of the items
    dates = []
    tile_ids = []
    for item in items:
        dates.append(item.properties["datetime_nominal"].date())
        tile_ids.append(item.properties["grid:code"])
    dates = np.unique(dates)
    tile_ids = np.unique(tile_ids)

    # sort items by date and tile ID into a data array
    grouped_items = np.full((len(dates), len(tile_ids)), None, dtype=object)
    for item in items:
        date = item.properties["datetime_nominal"].date()
        tile_id = item.properties["grid:code"]
        idx_date = np.where(dates == date)[0][0]
        idx_tile_id = np.where(tile_ids == tile_id)[0][0]
        if grouped_items[idx_date, idx_tile_id] is None:
            grouped_items[idx_date, idx_tile_id] = [item]
        else:
            grouped_items[idx_date, idx_tile_id].append(item)
    for idx_date in range(grouped_items.shape[0]):
        for idx_tile_id in range(grouped_items.shape[1]):
            if grouped_items[idx_date, idx_tile_id] is None:
                grouped_items[idx_date, idx_tile_id] = []
    grouped_items = xr.DataArray(
        grouped_items,
        dims=("time", "tile_id"),
        coords=dict(time=dates, tile_id=tile_ids),
    )

    # replace date by datetime from first item
    dts = []
    for date in grouped_items.time.values:
        item = np.sum(grouped_items.sel(time=date).values)[0]
        dts.append(item.datetime.replace(tzinfo=None))
    grouped_items = grouped_items.assign_coords(
        time=np.array(dts, dtype="datetime64[ns]")
    )
    grouped_items["time"].encoding["units"] = "seconds since 1970-01-01"
    grouped_items["time"].encoding["calendar"] = "standard"

    return grouped_items


def generate_cube(grouped_items: xr.DataArray, **open_params) -> xr.Dataset:
    """Generate a spatiotemporal data cube from grouped STAC items.

    This function takes grouped STAC items and generates a unified xarray
    dataset, mosaicking and stacking the data across spatial tiles and time.

    Args:
        grouped_items: A 2D data array indexed by time and tile_id, where each element
            contains a list of STAC items for the given date and tile ID.
        **open_params: Optional keyword arguments for data opening and processing:
            - bbox (list): Bounding box used for spatial subsetting.
            - crs (str): Coordinate reference system of the bounding box.
            - spatial_res (int | float | tuple): spatial resolution of the
                final datacube.

    Returns:
        A dataset representing the spatiotemporal cube.
    """
    # Group the tile IDs by UTM zones
    utm_tile_id = defaultdict(list)
    for tile_id in grouped_items.tile_id.values:
        item = np.sum(grouped_items.sel(tile_id=tile_id).values)[0]
        crs = item.properties["proj:code"]
        utm_tile_id[crs].append(tile_id)

    # Insert the tile data per UTM zone
    list_ds_utm = []
    for crs, tile_ids in utm_tile_id.items():
        ds = _generate_utm_cube(grouped_items.sel(tile_id=tile_ids), crs, **open_params)
        list_ds_utm.append(ds)

    # Reproject datasets from different UTM zones to a common grid reference system
    # and merge them into a single unified dataset for seamless spatial analysis.
    ds_final = _merge_utm_zones(list_ds_utm, **open_params)

    return ds_final


def _generate_utm_cube(
    grouped_items: xr.DataArray,
    crs_utm: str,
    **open_params,
) -> xr.Dataset:
    """Load and insert tile data for a specific UTM zone into a unified dataset.

    This method processes the assets from selected STAC items within a particular
    UTM zone. It opens the respective tiles, clips it to the target bounding box
    (reprojected to the UTM CRS), mosaics overlapping spatial datasets by taking the
    first valid pixel, and inserts the processed data into an aggregated dataset for
    that UTM zone.

    Args:
        grouped_items: A 2D data array indexed by time and tile_id, where each element
            contains a list of STAC items for the given date and tile ID.
        crs_utm: The target UTM coordinate reference system identifier.
        **open_params: Additional parameters to control data opening and processing,
            including 'bbox' (bounding box) and 'crs' (coordinate reference system).

    Returns:
        A dataset containing mosaicked and stacked 3d datacubes
        (time, spatial_y, spatial_x) for all requested spectral bands within the
        specified UTM zone.
    """
    items_bbox = _get_bounding_box(grouped_items)
    final_bbox = reproject_bbox(open_params["bbox"], open_params["crs"], crs_utm)
    spatial_res = _get_spatial_res(open_params)
    xarray_open_params = dict(
        resolution=spatial_res,
        interp_methods=open_params.get("interp_method"),
        agg_methods=open_params.get("agg_methods"),
        variables=open_params.get("variables"),
    )

    final_ds = None
    for dt_idx, dt in enumerate(grouped_items.time.values):
        for tile_id in grouped_items.tile_id.values:
            items = grouped_items.sel(tile_id=tile_id, time=dt).item()
            multi_tiles = []
            for item in items:
                ds = xr.open_dataset(
                    item.assets["product"].href,
                    engine="eopf-zarr",
                    chunks={},
                    **xarray_open_params,
                )
                ds = ds.sel(
                    x=slice(final_bbox[0], final_bbox[2]),
                    y=slice(final_bbox[3], final_bbox[1]),
                )
                if any(size == 0 for size in ds.sizes.values()):
                    continue
                multi_tiles.append(ds)
            if not multi_tiles:
                continue
            mosaicked_ds = mosaic_spatial_take_first(multi_tiles)
            if final_ds is None:
                final_ds = _create_empty_dataset(
                    mosaicked_ds, grouped_items, items_bbox, final_bbox, spatial_res
                )
            final_ds = _insert_tile_data(final_ds, mosaicked_ds, dt_idx)

    return final_ds


def _insert_tile_data(final_ds: xr.Dataset, ds: xr.Dataset, dt_idx: int) -> xr.Dataset:
    """Insert spatial data from a smaller dataset into a larger asset dataset at
    the correct spatiotemporal indices.

    This method locates the spatial coordinates of the input dataset `ds` within
    the larger `final_ds` along the 'x' and 'y' dimensions, then inserts the data
    from `ds` into the corresponding slice of `final_ds` for the specified time
    index `dt_idx`.

    Args:
        final_ds: The larger dataset representing the final data cube for one UTM zone.
        ds: The smaller xarray Dataset containing data to be inserted.
        dt_idx: The time index in `final_ds` at which to insert the data.

    Returns:
        The updated `final_ds` with data from `ds` inserted at the appropriate
        spatial location.
    """
    xmin = final_ds.indexes["x"].get_loc(ds.x[0].item())
    xmax = final_ds.indexes["x"].get_loc(ds.x[-1].item())
    ymin = final_ds.indexes["y"].get_loc(ds.y[0].item())
    ymax = final_ds.indexes["y"].get_loc(ds.y[-1].item())
    for var in ds.data_vars:
        final_ds[var][dt_idx, ymin : ymax + 1, xmin : xmax + 1] = ds[var]
    return final_ds


def _merge_utm_zones(list_ds_utm: list[xr.Dataset], **open_params) -> xr.Dataset:
    """Merge multiple Sentinel-2 datacubes for different UTM zones into a
    single dataset.

    This function takes a list of Sentinel-2 datasets (each in a different UTM zone),
    resamples them to a common grid defined by a target CRS and spatial resolution,
    and mosaics them into a single output using a "take first" strategy for overlaps.

    Parameters:
        list_ds_utm: A list of xarray Datasets, one for each UTM zone.
        open_params: Dictionary of parameters required for constructing the target grid,
            including:
            - crs: Target coordinate reference system (string or EPSG code).
            - spatial_res: Spatial resolution as a single value or tuple (x_res, y_res).
            - bbox: Bounding box for the output grid (minx, miny, maxx, maxy).
            - tile_size (optional): Tile size for the target grid.

    Returns:
        A single xarray Dataset reprojected to the target CRS and resolution,
        containing merged data from all input UTM zones.

    Notes:
        - If one input dataset already matches the target CRS and resolution,
          its grid mapping is reused unless resolution mismatches are found.
        - Overlapping regions are resolved by selecting the first non-NaN value.
    """
    # get correct target gridmapping
    crss = [pyproj.CRS.from_cf(ds["spatial_ref"].attrs) for ds in list_ds_utm]
    target_crs = pyproj.CRS.from_string(open_params["crs"])
    crss_equal = [target_crs == crs for crs in crss]
    if any(crss_equal):
        true_index = crss_equal.index(True)
        ds = list_ds_utm[true_index]
        target_gm = GridMapping.from_dataset(ds)
        spatial_res = open_params["spatial_res"]
        if not isinstance(spatial_res, tuple):
            spatial_res = (spatial_res, spatial_res)
        if (
            ds.x[1] - ds.x[0] != spatial_res[0]
            or abs(ds.y[1] - ds.y[0]) != spatial_res[1]
        ):
            target_gm = get_gridmapping(
                open_params["bbox"],
                open_params["spatial_res"],
                open_params["crs"],
                open_params.get("tile_size", _TILE_SIZE),
            )
    else:
        target_gm = get_gridmapping(
            open_params["bbox"],
            open_params["spatial_res"],
            open_params["crs"],
            open_params.get("tile_size", _TILE_SIZE),
        )

    resampled_list_ds = []
    for ds in list_ds_utm:
        resampled_list_ds.append(_resample_dataset_soft(ds, target_gm, **open_params))
    return mosaic_spatial_take_first(resampled_list_ds)


def _resample_dataset_soft(
    ds: xr.Dataset, target_gm: GridMapping, **open_params
) -> xr.Dataset:
    """Resample a dataset to a target grid mapping, using either affine transform
    or reprojection.

    If the source and target grid mappings are close, the dataset is returned unchanged.
    If they share the same CRS but differ spatially, an affine transformation is
    applied. Otherwise, the dataset is reprojected to the target grid.

    Parameters:
        ds: The source xarray Dataset to be resampled.
        target_gm: The target grid mapping

    Returns:
        The resampled dataset aligned with the target grid mapping.
    """
    source_gm = GridMapping.from_dataset(ds)
    if source_gm.is_close(target_gm):
        return ds
    ds = resample_in_space(
        ds,
        target_gm=target_gm,
        source_gm=source_gm,
        interp_methods=open_params.get("interp_methods"),
        agg_methods=open_params.get("agg_methods"),
    )
    return ds


def _get_bounding_box(grouped_items: xr.DataArray) -> list[float | int]:
    """Compute the overall bounding box that covers all tiles for the given grouped
     STAC items.

    Iterates through each tile ID in `grouped_items` to extract the bounding box
    from its STAC item and calculates the minimum bounding rectangle encompassing all
    tiles.

    Parameters:
        grouped_items: A 2D data array indexed by time and tile_id, where each element
            contains a list of STAC items for the given date and tile ID.

    Returns:
        A list with four elements [xmin, ymin, xmax, ymax] representing the
        bounding box that encloses all tiles.
    """
    xmin, ymin, xmax, ymax = np.inf, np.inf, -np.inf, -np.inf
    for tile_id in grouped_items.tile_id.values:
        item = np.sum(grouped_items.sel(tile_id=tile_id).values)[0]
        # take an assets which is available in L1C and L2A and read out
        # the bounding box in UTM coordinates.
        bbox = item.assets["B02_10m"].extra_fields["proj:bbox"]
        if xmin > bbox[0]:
            xmin = bbox[0]
        if ymin > bbox[1]:
            ymin = bbox[1]
        if xmax < bbox[2]:
            xmax = bbox[2]
        if ymax < bbox[3]:
            ymax = bbox[3]
    return [xmin, ymin, xmax, ymax]


def _get_spatial_res(open_params: dict) -> int:
    """Determine the appropriate Sentinel-2 spatial resolution based on the CRS.

    If the CRS is geographic (e.g., EPSG:4326), the requested spatial resolution
    (in degree) is converted to meters. The function then selects the nearest
    equal or coarser supported Sentinel-2 resolution from the native
    resolutions [10, 20, 60].

    Args:
        open_params: Dictionary of open parameters. Must include:
            - "crs": Coordinate reference system as EPSG string or identifier.
            - "spatial_res": Desired spatial resolution in meters.

    Returns:
        An integer representing the selected spatial resolution (10, 20, or 60).
        If no coarser resolution is found, defaults to 60.
    """
    crs = normalize_crs(open_params["crs"])
    if crs.is_geographic:
        spatial_res = open_params["spatial_res"] * CONVERSION_FACTOR_DEG_METER
    else:
        spatial_res = open_params["spatial_res"]
    idxs = np.where(_SEN2_SPATIAL_RES >= spatial_res)[0]
    if len(idxs) == 0:
        spatial_res = 60
    else:
        spatial_res = int(_SEN2_SPATIAL_RES[idxs[0]])

    return spatial_res


def _create_empty_dataset(
    sample_ds: xr.Dataset,
    grouped_items: xr.DataArray,
    items_bbox: list[float | int] | tuple[float | int],
    final_bbox: list[float | int] | tuple[float | int],
    spatial_res: int | float,
) -> xr.Dataset:
    """Create an empty xarray Dataset with spatial and temporal dimensions matching
    the given bounding boxes and grouped items.

    The dataset is constructed using the data variables and types from `sample_ds`,
    creating arrays filled with NaNs. It conforms to the native pixel grid and spatial
    resolution of the Sentinel-2 product, while covering the spatial extent defined
    by the input bounding boxes. The temporal dimension and coordinate values are
    derived from `grouped_items`. The resulting dataset includes coordinates for
    time, y, and x dimensions, along with a matching spatial reference coordinate
    system.

    Args:
        sample_ds: A sample dataset whose data variable names and dtypes will be used.
        grouped_items: A 2D DataArray (time, tile_id) containing grouped STAC items.
        items_bbox: The bounding box covering all input items (minx, miny, maxx, maxy).
        final_bbox: The target bounding box to define the spatial extent of the final
            datacube (minx, miny, maxx, maxy).
        spatial_res: The spatial resolution in CRS units (e.g., meters or degrees).

    Returns:
        A dataset with shape (time, y, x), filled with NaNs and ready to be populated
        with mosaicked data.
    """
    half_res = spatial_res / 2
    y_start = items_bbox[3] - spatial_res * (
        (items_bbox[3] - final_bbox[3]) // spatial_res
    )
    y_end = items_bbox[1] + spatial_res * (
        (final_bbox[1] - items_bbox[1]) // spatial_res
    )
    y = np.arange(y_start - half_res, y_end, -spatial_res)
    x_end = items_bbox[2] - spatial_res * (
        (items_bbox[2] - final_bbox[2]) // spatial_res
    )
    x_start = items_bbox[0] + spatial_res * (
        (final_bbox[0] - items_bbox[0]) // spatial_res
    )
    x = np.arange(x_start + half_res, x_end, spatial_res)

    chunks = (1, _TILE_SIZE, _TILE_SIZE)
    shape = (grouped_items.sizes["time"], len(y), len(x))
    ds = xr.Dataset(
        {
            key: (
                ("time", "y", "x"),
                da.full(shape, np.nan, dtype=var.dtype, chunks=chunks),
            )
            for (key, var) in sample_ds.data_vars.items()
        },
        coords={
            "x": x,
            "y": y,
            "time": grouped_items.time,
            "spatial_ref": sample_ds.spatial_ref,
        },
    )
    for key in sample_ds.data_vars:
        ds[key].attrs = {
            k: sample_ds[key].attrs[k]
            for k in _ATTRIBUTE_KEYS
            if k in sample_ds[key].attrs
        }

    return ds



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.


from abc import ABC, abstractmethod
from typing import Optional, Type

import pystac
import xarray as xr
from xcube.util.jsonschema import JsonObjectSchema


class ProductHandler(ABC):
    """Provides product-type specific properties and behaviour
    for the xcube EOPF-Zarr data store.
    """

    # Registry for product handler
    registry: "ProductHandlerRegistry"

    # data ID in this data store which represents the STAC Collection ID
    data_id: str

    @classmethod
    def guess(cls, data_id: str) -> Optional["ProductHandler"]:
        """Guess the suitable product handler for a given data id.

        Args:
            data_id: Data identifier of the data source.

        Returns:
            The product handler.

        Raises:
            ValueError: if guessing the product handler failed.
        """
        product_handler = ProductHandler.from_data_id(data_id)
        if product_handler is None:
            raise ValueError(
                f"Unable to detect product handler for {data_id!r}."
                " Use `data_id` argument to pass one of"
                f" {', '.join(map(repr, cls.registry.keys()))}."
            )
        return product_handler

    @classmethod
    def from_data_id(cls, data_id: str) -> Optional["ProductHandler"]:
        """Get the product handler for given `data_id`."""
        return cls.registry.get(data_id)

    @abstractmethod
    def get_open_data_params_schema(self) -> JsonObjectSchema:
        """Return opening parameters specific for the product handler."""

    @abstractmethod
    def open_data(self, items: list[pystac.Item], **open_params) -> xr.Dataset:
        """Open and return the dataset corresponding to the given parameters
        for the product handler.
        """


class ProductHandlerRegistry:
    """A simple registry for `ProductHandler` instances."""

    def __init__(self):
        self._product_handlers: dict[str, ProductHandler] = {}

    def keys(self) -> tuple[str, ...]:
        """Get registered product handler keys."""
        return tuple(self._product_handlers.keys())

    def values(self) -> tuple[ProductHandler, ...]:
        """Get registered product handlers."""
        # noinspection PyTypeChecker
        return tuple(self._product_handlers.values())

    def get(self, data_id: str) -> Optional["ProductHandler"]:
        """Get a specific product handler for given `data_id`."""
        return self._product_handlers.get(data_id)

    def register(self, cls: Type[ProductHandler]):
        """Register the product handler given as its class `cls`."""
        assert issubclass(cls, ProductHandler)
        assert isinstance(cls.data_id, str)
        self._product_handlers[cls.data_id] = cls()

    def unregister(self, cls: Type[ProductHandler]):
        """Unregister the product handler given as its class `cls`."""
        assert issubclass(cls, ProductHandler)
        assert isinstance(cls.data_id, str)
        del self._product_handlers[cls.data_id]


ProductHandler.registry = ProductHandlerRegistry()



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import dask.array as da
import numpy as np
import pyproj
import xarray as xr


def _sen2_sample(
    origin: tuple[int, int],
    size: tuple[int, int],
    resolution: int,
    chunksize: tuple[int, int],
    bands: list[str],
    crs_utm: pyproj.CRS,
):
    # Coordinate arrays (center of pixels)
    x = origin[0] + np.arange(size[1]) * resolution + resolution / 2
    y = origin[1] - np.arange(size[0]) * resolution - resolution / 2

    # Create dataset
    data_vars = {
        band: (
            ("y", "x"),
            da.zeros(size, chunks=chunksize, dtype=np.float64),
        )
        for band in bands
    }
    ds = xr.Dataset(
        data_vars=data_vars, coords={"x": ("x", x), "y": ("y", y), "spatial_ref": 0}
    )
    ds.coords["spatial_ref"].attrs = crs_utm.to_cf()
    if "scl" in ds:
        ds["scl"] = ds["scl"].astype(np.uint8)

    return ds


def sen2_l2a_10m() -> xr.Dataset:
    size = (10980, 10980)
    chunksize = (1830, 1830)
    resolution = 10
    origin = (600000, 5910000)
    crs_utm = pyproj.CRS.from_epsg(32632)
    bands = ["b02", "b03", "b04", "scl"]
    return _sen2_sample(origin, size, resolution, chunksize, bands, crs_utm)


def sen2_l2a_60m() -> xr.Dataset:
    size = (1830, 1830)
    chunksize = (305, 305)
    resolution = 60
    origin = (600000, 5910000)
    crs_utm = pyproj.CRS.from_epsg(32632)
    bands = ["b02", "b03", "b04", "scl"]
    return _sen2_sample(origin, size, resolution, chunksize, bands, crs_utm)


def sen2_l2a_60m_wo_scl() -> xr.Dataset:
    size = (1830, 1830)
    chunksize = (305, 305)
    resolution = 60
    origin = (600000, 5910000)
    crs_utm = pyproj.CRS.from_epsg(32632)
    bands = ["b02", "b03", "b04"]
    return _sen2_sample(origin, size, resolution, chunksize, bands, crs_utm)



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import datetime
from collections.abc import Sequence

import dask.array as da
import numpy as np
import pyproj
import pystac
import xarray as xr
from xcube.core.store import DataStoreError
from xcube_resampling.gridmapping import GridMapping

from .constants import STAC_URL
from .version import version


def reproject_bbox(
    source_bbox: Sequence[int | float],
    source_crs: pyproj.CRS | str,
    target_crs: pyproj.CRS | str,
    buffer: float = 0.0,
) -> Sequence[int | float]:
    """Reprojects a bounding box from a source CRS to a target CRS, with optional
    buffering.

    The function transforms a bounding box defined in the source coordinate reference
    system (CRS) to the target CRS using `pyproj`. If the source and target CRS are
    the same, no transformation is performed. An optional buffer (as a fraction of
    width/height) can be applied to expand the resulting bounding box.

    Args:
        source_bbox: The bounding box to reproject, in the form
            (min_x, min_y, max_x, max_y).
        source_crs: The source CRS, as a `pyproj.CRS` or string.
        target_crs: The target CRS, as a `pyproj.CRS` or string.
        buffer: Optional buffer to apply to the transformed bounding box, expressed as
                a fraction (e.g., 0.1 for 10% padding). Default is 0.0 (no buffer).

    Returns:
        A tuple representing the reprojected (and optionally buffered) bounding box:
        (min_x, min_y, max_x, max_y).
    """
    source_crs = normalize_crs(source_crs)
    target_crs = normalize_crs(target_crs)
    if source_crs != target_crs:
        t = pyproj.Transformer.from_crs(source_crs, target_crs, always_xy=True)
        target_bbox = t.transform_bounds(*source_bbox, densify_pts=21)
    else:
        target_bbox = source_bbox
    if buffer > 0.0:
        x_min = target_bbox[0]
        x_max = target_bbox[2]
        if target_crs.is_geographic and x_min > x_max:
            x_max += 360
        buffer_x = abs(x_max - x_min) * buffer
        buffer_y = abs(target_bbox[3] - target_bbox[1]) * buffer
        target_bbox = (
            target_bbox[0] - buffer_x,
            target_bbox[1] - buffer_y,
            target_bbox[2] + buffer_x,
            target_bbox[3] + buffer_y,
        )

    return target_bbox


def normalize_crs(crs: str | pyproj.CRS) -> pyproj.CRS:
    """Normalizes a CRS input by converting it to a pyproj.CRS object.

    If the input is already a `pyproj.CRS` instance, it is returned unchanged.
    If the input is a string (e.g., an EPSG code or PROJ string), it is converted
    to a `pyproj.CRS` object using `CRS.from_string`.

    Args:
        crs: A CRS specified as a string or a `pyproj.CRS` object.

    Returns:
        A `pyproj.CRS` object representing the normalized CRS.
    """
    if isinstance(crs, pyproj.CRS):
        return crs
    else:
        return pyproj.CRS.from_string(crs)


def add_nominal_datetime(items: Sequence[pystac.Item]) -> Sequence[pystac.Item]:
    """Adds the nominal (solar) time to each STAC item's properties under the key
    "datetime_nominal", based on the item's original UTC datetime.

    Args:
        items: A list of STAC item objects.

    Returns:
        A list of STAC item objects with the "datetime_nominal" field added to their
        properties.
    """

    for item in items:
        item.properties["center_point"] = get_center_from_bbox(item.bbox)
        item.properties["datetime_nominal"] = convert_to_solar_time(
            item.datetime, item.properties["center_point"][0]
        )
    return items


def get_center_from_bbox(bbox: Sequence[int | float]) -> Sequence[int | float]:
    """Calculates the center point of a bounding box.

    Args:
        bbox: The bounding box, in the form (min_x, min_y, max_x, max_y).

    Returns:
        A tuple (center_x, center_y) representing the center coordinates of the
        bounding box.
    """
    return (bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2


def convert_to_solar_time(
    utc: datetime.datetime, longitude: float
) -> datetime.datetime:
    """Converts a UTC datetime to an approximate solar time based on longitude.

    The conversion assumes that each 15 degrees of longitude corresponds to a 1-hour
    offset from UTC, effectively snapping the time offset to whole-hour increments.
    This provides a simplified approximation of local solar time.

    Args:
        utc: The datetime in UTC.
        longitude: The longitude in degrees, where positive values are east of
        the meridian.

    Returns:
        A datetime object representing the approximate solar time.
    """
    offset_seconds = int(longitude / 15) * 3600
    return utc + datetime.timedelta(seconds=offset_seconds)


def mosaic_spatial_take_first(list_ds: list[xr.Dataset]) -> xr.Dataset:
    """Creates a spatial mosaic from a list of datasets by taking the first
    non-NaN value encountered across datasets at each pixel location.

    The function assumes all datasets share the same spatial dimensions and coordinate
    system. Only variables with 2D spatial dimensions (y, x) are processed. At each
    spatial location, the first non-NaN value across the dataset stack
    is selected.

    Args:
        list_ds: A list of datasets to be mosaicked.

    Returns:
        A new dataset representing the mosaicked result, using the first valid
        value encountered across the input datasets for each spatial position.
    """
    if len(list_ds) == 1:
        return list_ds[0]

    y_coord, x_coord = get_spatial_dims(list_ds[0])
    ds_mosaic = xr.Dataset()
    for key in list_ds[0]:
        if list_ds[0][key].dims[-2:] == (y_coord, x_coord):
            da_arr = da.stack([ds[key].data for ds in list_ds], axis=0)
            nonnan_mask = ~da.isnan(da_arr)
            first_non_nan_index = nonnan_mask.argmax(axis=0)
            da_arr_select = da.choose(first_non_nan_index, da_arr)
            ds_mosaic[key] = xr.DataArray(
                da_arr_select,
                dims=list_ds[0][key].dims,
                coords=list_ds[0][key].coords,
                attrs=list_ds[0][key].attrs,
            )

    # attributes are taken from the first UTM dataset
    ds_mosaic.attrs = list_ds[0].attrs

    return ds_mosaic


def get_spatial_dims(ds: xr.Dataset) -> (str, str):
    """Identifies the spatial coordinate names in a dataset.

    The function checks for common spatial dimension naming conventions: ("lat", "lon")
    or ("y", "x"). If neither pair is found, it raises a DataStoreError.

    Args:
        ds: The dataset to inspect.

    Returns:
        A tuple of strings representing the names of the spatial dimensions.

    Raises:
        DataStoreError: If no recognizable spatial dimensions are found.
    """
    if "lat" in ds and "lon" in ds:
        y_coord, x_coord = "lat", "lon"
    elif "y" in ds and "x" in ds:
        y_coord, x_coord = "y", "x"
    else:
        raise DataStoreError("No spatial dimensions found in dataset.")
    return y_coord, x_coord


def get_gridmapping(
    bbox: get_spatial_dims,
    spatial_res: float | int | tuple[float | int, float | int],
    crs: str | pyproj.crs.CRS,
    tile_size: int | tuple[int, int],
) -> GridMapping:
    """Creates a regular GridMapping object based on a bounding box, spatial resolution,
    and CRS.

    Args:
        bbox: The bounding box in the form (min_x, min_y, max_x, max_y).
        spatial_res: Spatial resolution as a single value or a (x_res, y_res) tuple.
        crs: Coordinate reference system as a `pyproj.CRS` or a string
            (e.g., "EPSG:4326").
        tile_size: Tile size as a single integer or a (width, height) tuple.

    Returns:
        A xcube `GridMapping` object representing the regular grid layout defined
        by the input parameters.
    """
    if not isinstance(spatial_res, tuple):
        spatial_res = (spatial_res, spatial_res)
    x_size = np.ceil((bbox[2] - bbox[0]) / spatial_res[0]) + 1
    y_size = np.ceil(abs(bbox[3] - bbox[1]) / spatial_res[1]) + 1
    return GridMapping.regular(
        size=(x_size, y_size),
        xy_min=(bbox[0] - spatial_res[0] / 2, bbox[1] - spatial_res[1] / 2),
        xy_res=spatial_res,
        crs=crs,
        tile_size=tile_size,
    )


def add_attributes(
    ds: xr.Dataset, grouped_items: xr.DataArray, **open_params
) -> xr.Dataset:
    """Adds metadata attributes to the final dataset.

    This function enriches the input dataset with additional metadata attributes:
    - 'stac_url': A predefined URL for the EOPF STAC API.
    - 'stac_items': A mapping from time steps to lists of STAC item IDs.
    - 'open_params': Opening parameters.
    - 'xcube_eopf_version': The version of the xcube-eopf package being used.

    Parameters:
        ds: The input dataset to which attributes will be added.
        grouped_items: An array containing STAC items grouped by time and tile ID.
        **open_params: Opening parameters that are stored as a metadata attribute.

    Returns:
        The modified dataset with added metadata attributes.
    """
    ds.attrs["stac_url"] = STAC_URL
    ds.attrs["stac_items"] = dict(
        {
            dt.astype("datetime64[ms]")
            .astype("O")
            .isoformat(): [
                item.id for item in np.sum(grouped_items.sel(time=dt).values)
            ]
            for dt in grouped_items.time.values
        }
    )
    ds.attrs["open_params"] = open_params
    ds.attrs["xcube_eopf_version"] = version

    return ds


def filter_items_deprecated(items: list[pystac.Item]) -> list[pystac.Item]:
    """Filter out deprecated STAC items, which are deprecated.

    Args:
        items: A list of STAC items to filter.

    Returns:
        A list of STAC items that are not marked as deprecated.
    """
    sel_items = []
    for item in items:
        deprecated = item.properties.get("deprecated", False)
        if not deprecated:
            sel_items.append(item)
    return sel_items


def filter_items_wrong_footprint(items: list[pystac.Item]) -> list[pystac.Item]:
    """Filter out STAC items with incorrectly assigned footprints at the antimeridian.

    Some items may have footprints crossing the antimeridian where the west and east
    boundaries are swapped, resulting in a longitude span greater than 180°.
    This function removes such items by checking the difference between the minimum
    and maximum longitude values in the item's bounding box.

    Args:
        items: A list of STAC items to filter.

    Returns:
        A list of STAC items whose bounding boxes do not exceed 180°
        in longitude extent.

    Notes:
        See related issue: https://github.com/EOPF-Sample-Service/eopf-stac/issues/39
    """
    sel_items = []
    for item in items:
        if abs(item.bbox[2] - item.bbox[0]) < 180:
            sel_items.append(item)
    return sel_items


def bbox_to_geojson(bbox):
    """
    Convert a bounding box to a GeoJSON Polygon.

    Args:
        bbox: list or tuple of four numbers [min_x, min_y, max_x, max_y]

    Returns:
        dict: GeoJSON Polygon
    """
    min_x, min_y, max_x, max_y = bbox
    return {
        "type": "Polygon",
        "coordinates": [
            [
                [min_x, min_y],
                [max_x, min_y],
                [max_x, max_y],
                [min_x, max_y],
                [min_x, min_y],
            ]
        ],
    }



#  Copyright (c) 2025 by EOPF Sample Service team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from .version import version as __version__

__all__ = ["__version__"]
